==> wurld-0.1.tar.gz: wurld-0.1/src/GameEngine.cc <==

#include "GameEngine.h"
#include <assert.h>
#include <spug/FileReader.h>
#include "Scene.h"
#include "NativePersistor.h"
#include "Color.h"
#include "ColorMaterial.h"
#include "PhysicalObject.h"
#include "Light.h"
#include "Graphics.h"

#include "QuadObj.h" // needed by bogus makeCube()
#include <GL/glu.h>

using namespace spug;
using namespace std;

GameEngine::GameEngine() : 
   forward(false), 
   backward(false), 
   left(false), 
   right(false) {

   // initialize the camera origin
   camera.setOrigin(Vector(0, 18, 0));

   // initialize the camera rotation matrix
   RotMatrix initialRotation;
   initialRotation.makeIdent();
   camera.setRotation(initialRotation);

   // enable detailed tracing of the persistor
   //Tracer::get().enable(Tracer::get().getLevel("NativePersistor.detailed"));

   // initialize the frame clock
   gettimeofday(&lastEngineFrame, 0);
   lastRenderFrame = lastEngineFrame;

   // initialize the player - he's special (for now) because he defines the
   // camera position.
   player = new PhysicalObject();
   player->setPosition(camera.getOrigin());
   player->setVelocity(Vector(0, 0, 0));

   // add the player to the game objects
   objects.push_back(GameObjectPtr::ucast(player));
}

void GameEngine::loadMap(const char *mapName) {
   // load the map
   try {
      FileReader mapSrc("map.mdl");
      NativePersistor pers;
      pers.readModel(mapSrc, map);
      mapSrc.close();
   } catch (const NativePersistor::OutOfDataError &ex) {
      throw InvalidFileFormatError("map.mdl");
   } catch (const Exception &ex) {
      throw FileDoesNotExistError("map.mdl");
   }

   // get the map's bounding box and use it to calculate the maximum distance
   // we will ever have to see
   AxisBox perimeter = map.getBoundingBox();
   maxDistance = (perimeter.max - perimeter.min).size();
}

void GameEngine::loadTextures() {

// XXX we are demand-loading textures for now, revisit this later
#if 0
   try {
      // load the texture
      NativePersistor pers;
      pers.loadTexture("stucco.png");
   } catch (const Exception &ex) {
      cout << ex << endl;
   } catch (...) {
      cout << "Got anonymous exception" << endl;
   }
#endif
}

void GameEngine::processEngineFrame() {

   // move the player

   // create the direction vector relative to the camera orientation, then
   // rotate it to convert it to an absolute vector.
   Vector 
      relDirVec((left ? -1 : 0) + (right ? 1 : 0), 
		0, 
		(forward ? -1 : 0) + (backward ? 1 : 0)
		), 
      absDirVec;
   if (relDirVec.x || relDirVec.z) {
      camera.getRotation().rotate(relDirVec, absDirVec);

      // remove the y component (want to stay grounded)
      absDirVec.y = 0;

      // normalize it for collision calculation
      absDirVec.normalize();

#if 0
      // calculate the distance we are going to move (velocity * seconds since
      // last frame)
      Float moveDistance = 50.0 * (diff / 1000.0);
#endif
      // set the player velocity to the desired direction
      player->setVelocity(absDirVec * 50);

      // check for collisions with the world.  
      // for now, just detect a ray collision  There are potentially many
      // collision points for ray, so we find the nearest one
      Model::PolyIter iter;
      Vector intersection, origin = camera.getOrigin();
      Float distance, maxDistance = infinity;
      for (Model::PolyIter iter = map.getPolyIterator(); iter; ++iter)
	 if ((*iter)->intersects(origin, absDirVec, intersection,
				 distance
				 ) &&
              distance < maxDistance
	     ) {
            maxDistance = distance;
	 }

      // limit the distance we will travel to slightly before the collision
      // point.  If we are already within that distance, reverse direction.
      if (maxDistance >= 2)
         player->setMaxDistance(maxDistance - 2);
      else {
         player->setMaxDistance(2);  // XXX should check for reverse collision
         player->setVelocity(absDirVec * -5);
      }
   } else {
      player->setVelocity(Vector(0, 0, 0));
   }

   // Call the processEngineFrame() method of all game objects
   for (GameObjectSequence::iterator iter = objects.begin(); 
	iter != objects.end();
	++iter
        )
      (*iter)->processEngineFrame(*this);
}

void GameEngine::processRenderFrame() {

   // determine whether 1/framesPerSecond seconds have elapsed since the last
   // frame

   // get the current time of day
   struct timeval now;
   gettimeofday(&now, NULL);

   // get the difference in seconds
   unsigned diff = now.tv_sec - lastEngineFrame.tv_sec;
   // if > 100 seconds since the last frame, we have a real problem here..
   assert(diff < 100);  

   // convert difference to difference in milliseconds
   diff *= 1000;
   diff += (now.tv_usec - lastEngineFrame.tv_usec) / 1000;

   // if we're due for an engine frame, do it now.
   if (diff > 1000/framesPerSecond) {
      engineDelta = diff / 1000.0;
      processEngineFrame();
      lastEngineFrame = now;
   }

   // calculate the time since the last render frame
   diff = (now.tv_sec - lastRenderFrame.tv_sec) * 1000 +
      (now.tv_usec - lastRenderFrame.tv_usec) / 1000;
   renderDelta = diff / 1000.0;

   // store the rendering frames-per-second
   renderFramesPerSecond = 1000000.0 / 
      ((now.tv_sec - lastRenderFrame.tv_sec) * 1000000 +
       (now.tv_usec - lastRenderFrame.tv_usec));

   lastRenderFrame = now;

   // perform all render frame processing
   for (GameObjectSequence::iterator iter = objects.begin();
        iter != objects.end();
	++iter
        )
      (*iter)->processRenderFrame(*this, renderDelta);

   // set the camera position from the player
   camera.setOrigin(player->getPosition());

   // render the scene
   render();
}

void GameEngine::render() {
   // render all of the sticky objects
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity( );

   /* Set our perspective */
   gluPerspective( 45.0f, static_cast<float>(640) / 480, 0.1f, 100.0f );

   glLoadIdentity();
   glTranslate<Float>(0, 0, -0.1);
   Graphics &grap = Graphics::getInstance();
   grap.setColor(Color(0, 0, 0));
   glColor3f(1, 0, 0);

   char fpsBuffer[20];
   sprintf(fpsBuffer, "FPS: %f", renderFramesPerSecond);
   grap.moveTo(-1, -1);
   grap.drawText(fpsBuffer);

   // render everything else
   Scene scene(camera.getOrigin(), camera.getRotation(), maxDistance);
   scene.mapToCamera();
   glLoadIdentity();

   // render the sticky objects
#if 0
   for obj in stickyObjects {

      // rotate the object's position so that it is static with respect to the
      // camera orientation
      Vector objPos;
      camera.getRotation().rotate(obj->getPosition(), objPos);
      buddy->setPosition(player->getPosition() + objPos);

      RotMatrix objOrientation;
      objOrientation.makeInverse(camera.getRotation());
      buddy->setOrientation(objOrientation);
      obj->render(scene);
   }
#endif

   // give the player a light - XXX should be another sticky object
   Light(player->getPosition(), Color(1.0, 1.0, 1.0)).render(scene);

   // render all objects
   for (RenderableObjectSequence::iterator iter = renderObjects.begin();
        iter != renderObjects.end();
	++iter
	)
      (*iter)->render(scene);

   // render the map
   map.render(scene);
}

void GameEngine::handleIdle() {
   processRenderFrame();
}

void GameEngine::addGameObject(const GameObjectPtr &object) {
   objects.push_back(object);
}

void GameEngine::addRenderableObject(const RenderableObjectPtr &object) {
   renderObjects.push_back(object);
}

// XXX hack - remove me
// initializes the "cube" global variable
void GameEngine::makeCube(Model &cube) {

   // initialize the cube
   Vector 
      frontLeftTop(-1, 1, -1),
      frontRightTop(1, 1, -1),
      frontRightBottom(1, -1, -1),
      frontLeftBottom(-1, -1, -1),
      backLeftTop(-1, 1, 1),
      backRightTop(1, 1, 1),
      backRightBottom(1, -1, 1),
      backLeftBottom(-1, -1, 1);

   MaterialPtr 
      red = new ColorMaterial(Color(1, 0, 0)), 
      green = new ColorMaterial(Color(0, 1, 0)), 
      blue = new ColorMaterial(Color(0, 0, 1)), 
      yellow = new ColorMaterial(Color(1, 1, 0)),
      cyan = new ColorMaterial(Color(0, 1, 1)), 
      magenta = new ColorMaterial(Color(1, 0, 1));

   // front
   cube.addPoly(
      new QuadObj(frontLeftTop, frontRightTop, frontRightBottom,
		  frontLeftBottom,
		  red
		  )
   );

   // top
   cube.addPoly(
      new QuadObj(frontLeftTop, backLeftTop, backRightTop,
		  frontRightTop,
		  green
		  )
   );

   // back
   cube.addPoly(
      new QuadObj(backLeftTop, backRightTop, backRightBottom,
		  backLeftBottom,
		  blue
		  )
   );

   // bottom
   cube.addPoly(
      new QuadObj(frontLeftBottom, backLeftBottom, backRightBottom,
		  frontRightBottom,
		  yellow
		  )
   );

   // right side
   cube.addPoly(
      new QuadObj(frontRightTop, backRightTop, backRightBottom,
		  frontRightBottom,
		  cyan
		  )
   );

   // left side
   cube.addPoly(
      new QuadObj(frontLeftTop, backLeftTop, backLeftBottom,
		  frontLeftBottom,
		  magenta
		  )
   );
}

==> wurld-0.2.tar.gz: wurld-0.2/src/GameEngine.cc <==

#include "GameEngine.h"
#include <assert.h>
#include <spug/FileReader.h>
#include "Scene.h"
#include "NativePersistor.h"
#include "Color.h"
#include "ColorMaterial.h"
#include "PhysicalObject.h"
#include "Light.h"
#include "Graphics.h"
#include "KeyCommand.h"

#include "QuadObj.h" // needed by bogus makeCube()
#include <GL/glu.h>

using namespace spug;
using namespace std;

//---------------------------------------------------
// All of the built-in key commands are defined here.
//---------------------------------------------------

namespace {
   using namespace wurld;

   class ForwardCmd : public KeyCommand {
      private:
         GameEngine &engine;
      public:
         ForwardCmd(GameEngine &engine) : engine(engine) {}
         virtual void handleKeyPress(KeyCode key) {
            engine.setForward(true);
         }
         virtual void handleKeyRelease(KeyCode key) {
            engine.setForward(false);
         }
   };

   class BackwardCmd : public KeyCommand {
      private:
         GameEngine &engine;
      public:
         BackwardCmd(GameEngine &engine) : engine(engine) {}
         virtual void handleKeyPress(KeyCode key) {
            engine.setBackward(true);
         }
         virtual void handleKeyRelease(KeyCode key) {
            engine.setBackward(false);
         }
   };

   class LeftCmd : public KeyCommand {
      private:
         GameEngine &engine;
      public:
         LeftCmd(GameEngine &engine) : engine(engine) {}
         virtual void handleKeyPress(KeyCode key) {
            engine.setLeft(true);
         }
         virtual void handleKeyRelease(KeyCode key) {
            engine.setLeft(false);
         }
   };

   class RightCmd : public KeyCommand {
      private:
         GameEngine &engine;
      public:
         RightCmd(GameEngine &engine) : engine(engine) {}
         virtual void handleKeyPress(KeyCode key) {
            engine.setRight(true);
         }
         virtual void handleKeyRelease(KeyCode key) {
            engine.setRight(false);
         }
   };
}

void GameEngine::initKeyCommands() {
   registerKeyCommand("forward", new ForwardCmd(*this));
   registerKeyCommand("backward", new BackwardCmd(*this));
   registerKeyCommand("left", new LeftCmd(*this));
   registerKeyCommand("right", new RightCmd(*this));
}

GameEngine::GameEngine() : 
   forward(false), 
   backward(false), 
   left(false), 
   right(false) {

   // initialize the camera origin
   camera.setOrigin(Vector(0, 18, 0));

   // initialize the camera rotation matrix
   RotMatrix initialRotation;
   initialRotation.makeIdent();
   camera.setRotation(initialRotation);

   // enable detailed tracing of the persistor
   //Tracer::get().enable(Tracer::get().getLevel("NativePersistor.detailed"));

   // initialize the frame clock
   gettimeofday(&lastEngineFrame, 0);
   lastRenderFrame = lastEngineFrame;

   // initialize the player - he's special (for now) because he defines the
   // camera position.
   player = new PhysicalObject();
   player->setPosition(camera.getOrigin());
   player->setVelocity(Vector(0, 0, 0));

   // add the player to the game objects
   objects.push_back(GameObjectPtr::ucast(player));

   // initialize all default key commands
   initKeyCommands();
}

void GameEngine::loadMap(const char *mapName) {
   // load the map
   try {
      FileReader mapSrc("map.mdl");
      NativePersistor pers;
      pers.readModel(mapSrc, map);
      mapSrc.close();
   } catch (const NativePersistor::OutOfDataError &ex) {
      throw InvalidFileFormatError("map.mdl");
   } catch (const Exception &ex) {
      throw FileDoesNotExistError("map.mdl");
   }

   // get the map's bounding box and use it to calculate the maximum distance
   // we will ever have to see
   AxisBox perimeter = map.getBoundingBox();
   maxDistance = (perimeter.max - perimeter.min).size();
}

void GameEngine::loadTextures() {

// XXX we are demand-loading textures for now, revisit this later
#if 0
   try {
      // load the texture
      NativePersistor pers;
      pers.loadTexture("stucco.png");
   } catch (const Exception &ex) {
      cout << ex << endl;
   } catch (...) {
      cout << "Got anonymous exception" << endl;
   }
#endif
}

void GameEngine::processEngineFrame() {

   // move the player

   // create the direction vector relative to the camera orientation, then
   // rotate it to convert it to an absolute vector.
   Vector 
      relDirVec((left ? -1 : 0) + (right ? 1 : 0), 
		0, 
		(forward ? -1 : 0) + (backward ? 1 : 0)
		), 
      absDirVec;
   if (relDirVec.x || relDirVec.z) {
      camera.getRotation().rotate(relDirVec, absDirVec);

      // remove the y component (want to stay grounded)
      absDirVec.y = 0;

      // normalize it for collision calculation
      absDirVec.normalize();

#if 0
      // calculate the distance we are going to move (velocity * seconds since
      // last frame)
      Float moveDistance = 50.0 * (diff / 1000.0);
#endif
      // set the player velocity to the desired direction
      player->setVelocity(absDirVec * 50);

      // check for collisions with the world.  
      // for now, just detect a ray collision  There are potentially many
      // collision points for ray, so we find the nearest one
      Model::PolyIter iter;
      Vector intersection, origin = camera.getOrigin();
      Float distance, maxDistance = infinity;
      for (Model::PolyIter iter = map.getPolyIterator(); iter; ++iter)
	 if ((*iter)->intersects(origin, absDirVec, intersection,
				 distance
				 ) &&
              distance < maxDistance
	     ) {
            maxDistance = distance;
	 }

      // limit the distance we will travel to slightly before the collision
      // point.  If we are already within that distance, reverse direction.
      if (maxDistance >= 2)
         player->setMaxDistance(maxDistance - 2);
      else {
         player->setMaxDistance(2);  // XXX should check for reverse collision
         player->setVelocity(absDirVec * -5);
      }
   } else {
      player->setVelocity(Vector(0, 0, 0));
   }

   // Call the processEngineFrame() method of all game objects
   for (GameObjectSequence::iterator iter = objects.begin(); 
	iter != objects.end();
	++iter
        )
      (*iter)->processEngineFrame(*this);
}

void GameEngine::processRenderFrame() {

   // determine whether 1/framesPerSecond seconds have elapsed since the last
   // frame

   // get the current time of day
   struct timeval now;
   gettimeofday(&now, NULL);

   // get the difference in seconds
   unsigned diff = now.tv_sec - lastEngineFrame.tv_sec;
   // if > 100 seconds since the last frame, we have a real problem here..
   assert(diff < 100);  

   // convert difference to difference in milliseconds
   diff *= 1000;
   diff += (now.tv_usec - lastEngineFrame.tv_usec) / 1000;

   // if we're due for an engine frame, do it now.
   if (diff > 1000/framesPerSecond) {
      engineDelta = diff / 1000.0;
      processEngineFrame();
      lastEngineFrame = now;
   }

   // calculate the time since the last render frame
   diff = (now.tv_sec - lastRenderFrame.tv_sec) * 1000 +
      (now.tv_usec - lastRenderFrame.tv_usec) / 1000;
   renderDelta = diff / 1000.0;

   // store the rendering frames-per-second
   renderFramesPerSecond = 1000000.0 / 
      ((now.tv_sec - lastRenderFrame.tv_sec) * 1000000 +
       (now.tv_usec - lastRenderFrame.tv_usec));

   lastRenderFrame = now;

   // perform all render frame processing
   for (GameObjectSequence::iterator iter = objects.begin();
        iter != objects.end();
	++iter
        )
      (*iter)->processRenderFrame(*this, renderDelta);

   // set the camera position from the player
   camera.setOrigin(player->getPosition());

   // render the scene
   render();
}

void GameEngine::render() {
   // render everything else
   Scene scene(camera.getOrigin(), camera.getRotation(), maxDistance);
   scene.mapToCamera();
   glLoadIdentity();

   // render the sticky objects
#if 0
   for obj in stickyObjects {

      // rotate the object's position so that it is static with respect to the
      // camera orientation
      Vector objPos;
      camera.getRotation().rotate(obj->getPosition(), objPos);
      buddy->setPosition(player->getPosition() + objPos);

      RotMatrix objOrientation;
      objOrientation.makeInverse(camera.getRotation());
      buddy->setOrientation(objOrientation);
      obj->render(scene);
   }
#endif

   // render all objects
   for (RenderableObjectSequence::iterator iter = renderObjects.begin();
        iter != renderObjects.end();
	++iter
	)
      (*iter)->render(scene);

   // render the map
   map.render(scene);

   // give the player a light - XXX should be another sticky object
   Light(player->getPosition(), Color(1.0, 1.0, 1.0)).render(scene);

   // render all of hud stickers - these always appear to float on the front
   // of the view

   // first, restore the projection matrix to its ortho-box mode so the text
   // doesn't get perspective
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity();
   glTranslate<Float>(0, 0, -0.1);

   // disable depth and lighting for the text portion of our rendering
   Graphics &grap = Graphics::getInstance();
   bool lighting = grap.lightingEnabled();
   glDisable(GL_DEPTH_TEST);
   if (lighting) glDisable(GL_LIGHTING);

   // draw the FPS
   grap.setColor(Color(0, 1, 0));

   // draw all of the other HUD stickers
   RenderableObjectSequence::iterator hudIter = hudStickers.begin();
   for (; hudIter != hudStickers.end(); ++hudIter)
      (*hudIter)->render(scene);

   glEnable(GL_DEPTH_TEST);
   if (lighting) glEnable(GL_LIGHTING);

}

void GameEngine::handleIdle() {
   processRenderFrame();
}

void GameEngine::addGameObject(const GameObjectPtr &object) {
   objects.push_back(object);
}

void GameEngine::addRenderableObject(const RenderableObjectPtr &object) {
   renderObjects.push_back(object);
}

void GameEngine::registerKeyCommand(const char *name, 
                                    const KeyCommandPtr &command
                                    ) {
   keyCommands[name] = command;
}

KeyCommandPtr GameEngine::getKeyCommand(const char *name) const {
   KeyCommandMap::const_iterator iter = keyCommands.find(name);
   if (iter != keyCommands.end())
      return iter->second;
   else
      return 0;
}

void GameEngine::addHUDSticker(const RenderableObjectPtr &hudSticker) {
   hudStickers.push_back(hudSticker);
}

void GameEngine::removeHUDSticker(const RenderableObjectPtr &hudSticker) {
   RenderableObjectSequence::iterator iter = hudStickers.begin();
   for (; *iter != hudSticker && iter != hudStickers.end(); ++iter) ;
   if (iter != hudStickers.end())
      hudStickers.erase(iter);
   else
      throw NotFoundError("HUD Object not found for removal");
}

// XXX hack - remove me
// initializes the "cube" global variable
void GameEngine::makeCube(Model &cube) {

   // initialize the cube
   Vector 
      frontLeftTop(-1, 1, -1),
      frontRightTop(1, 1, -1),
      frontRightBottom(1, -1, -1),
      frontLeftBottom(-1, -1, -1),
      backLeftTop(-1, 1, 1),
      backRightTop(1, 1, 1),
      backRightBottom(1, -1, 1),
      backLeftBottom(-1, -1, 1);

   MaterialPtr 
      red = new ColorMaterial(Color(1, 0, 0)), 
      green = new ColorMaterial(Color(0, 1, 0)), 
      blue = new ColorMaterial(Color(0, 0, 1)), 
      yellow = new ColorMaterial(Color(1, 1, 0)),
      cyan = new ColorMaterial(Color(0, 1, 1)), 
      magenta = new ColorMaterial(Color(1, 0, 1));

   // front
   cube.addPoly(
      new QuadObj(frontLeftTop, frontRightTop, frontRightBottom,
		  frontLeftBottom,
		  red
		  )
   );

   // top
   cube.addPoly(
      new QuadObj(frontLeftTop, backLeftTop, backRightTop,
		  frontRightTop,
		  green
		  )
   );

   // back
   cube.addPoly(
      new QuadObj(backLeftTop, backRightTop, backRightBottom,
		  backLeftBottom,
		  blue
		  )
   );

   // bottom
   cube.addPoly(
      new QuadObj(frontLeftBottom, backLeftBottom, backRightBottom,
		  frontRightBottom,
		  yellow
		  )
   );

   // right side
   cube.addPoly(
      new QuadObj(frontRightTop, backRightTop, backRightBottom,
		  frontRightBottom,
		  cyan
		  )
   );

   // left side
   cube.addPoly(
      new QuadObj(frontLeftTop, backLeftTop, backLeftBottom,
		  frontLeftBottom,
		  magenta
		  )
   );
}
